"""
Program to check tables included in precis_i18n.unicode module.
"""

import re
from urllib.request import urlopen

import precis_i18n.unicode as ucd
from precis_i18n.codepointset import CodepointSet

PROP_REGEX = re.compile(rb"^([0-9A-Za-z.]+)\s+;\s+(\w+)\s+#")
DATAFILE_URLS = [
    "https://www.unicode.org/Public/UNIDATA/DerivedCoreProperties.txt",
    "https://www.unicode.org/Public/UNIDATA/extracted/DerivedJoiningType.txt#Join_Type",
    "https://www.unicode.org/Public/UNIDATA/Scripts.txt",
    'https://www.unicode.org/Public/UNIDATA/HangulSyllableType.txt#Hangul_Type',
]

def parse_unicode_datafile(url, props):
    if '#' in url:
        url, prop = url.split("#")
        prop += '='
    else:
        prop = ''

    for line in urlopen(url):
        m = PROP_REGEX.match(line)
        if m:
            codepoints, name = m.groups()
            props.setdefault('%s%s' % (prop, name.decode()), []).append(codepoints.decode())


def compare_codepoints(varname, codepoint_str):
    codepoints = CodepointSet(codepoint_str)
    if codepoints != getattr(ucd, varname):
        print(varname)
        print(codepoint_str)


def combine(props, *names):
    result = ''
    for name in names:
        cnt = len(CodepointSet(props[name]))
        result += '# %s (%d)\n%s\n' % (name, cnt, props[name])
    return result


def main():
    props = {}
    for url in DATAFILE_URLS:
        parse_unicode_datafile(url, props)

    for prop in props:
        props[prop] = '\n'.join(props[prop])

    compare_codepoints('_DEFAULT_IGNORABLE', props['Default_Ignorable_Code_Point'])
    compare_codepoints('_JOINTYPE_DUAL_JOINING', props['Join_Type=D'])
    compare_codepoints('_JOINTYPE_RIGHT_JOINING', props['Join_Type=R'])
    compare_codepoints('_JOINTYPE_LEFT_JOINING', props['Join_Type=L'])
    compare_codepoints('_JOINTYPE_TRANSPARENT', props['Join_Type=T'])
    compare_codepoints('_GREEK_SCRIPT', props['Greek'])
    compare_codepoints('_HEBREW_SCRIPT', props['Hebrew'])
    compare_codepoints('_HIRAGANA_KATAKANA_HAN', combine(props, 'Hiragana', 'Katakana', 'Han'))
    compare_codepoints('_OLD_HANGUL_JAMO', combine(props, 'Hangul_Type=L', 'Hangul_Type=V', 'Hangul_Type=T'))


if __name__ == '__main__':
    main()



import os
import sys
import unittest

from precis_i18n.derived import derived_property
from precis_i18n.unicode import UnicodeData

try:
    import unicodedata2
except ImportError:
    unicodedata2 = None


class TestDerivedProperties(unittest.TestCase):
    """ Test output of derived_property function.
    """
    def test_derived_props(self):
        """ Compare derived properties against a "golden" file using built-in
        unicodedata module.
        """
        self._test_derived_props(UnicodeData())

    @unittest.skipIf(unicodedata2 is None, 'unicodedata2 not available')
    def test_derived_props_unicodedata2(self):
        """ Compare derived properties against a "golden" file using
        pip-installed unicodedata2 module.
        """

        ucd = UnicodeData(unicodedata2)
        assert ucd.version in  {10.0, 12.0, 14.0, 15.0}
        self._test_derived_props(ucd)

    def _test_derived_props(self, ucd):
        """ Compare derived properties against a "golden" file.
        """

        goldname = 'derived-props-%.1f.txt' % ucd.version
        goldpath = os.path.join(os.path.dirname(__file__), goldname)

        with open(goldpath) as golden:
            derived_props = enumerate_derived_props(ucd)
            for line in golden:
                line = line.strip()
                if not line:
                    continue
                expected = '%04X-%04X %s' % next(derived_props)
                self.assertEqual(line, expected)


def enumerate_derived_props(ucd, begin=0, end=0x00110000):
    """ Iterable producing all derived_property values for a particular range.

    Produces a sequence of tuples (begin, end, property).
    """
    start = begin
    prev = '/'.join(derived_property(start, ucd))
    for cp in range(start + 1, end):
        prop = '/'.join(derived_property(cp, ucd))
        if prop != prev:
            yield (start, cp - 1, prev)
            start = cp
            prev = prop
    yield (start, cp, prev)


def main():
    ucd = UnicodeData()
    print('Unicode %.1f' % ucd.version, file=sys.stderr)
    # Dump gold file to stdout.
    for p in enumerate_derived_props(ucd):
        print('%04X-%04X %s' % p)


if __name__ == '__main__':
    main()



# load_unicode.py

# Load unicode data into a sqlite database.
#
#  cp   codepoint
#  name name
#  cat  general_category
#  combining  combining_category
#  bidi

import re
import sqlite3


class UnicodeDatabase(object):

    property_regex = re.compile(
        r'^([0-9A-F]+)(?:\.\.([0-9A-F]+))?\s*;\s*(\S+)\s*[#;]')

    def __init__(self, filename):
        self._conn = sqlite3.connect(filename)

    def load(self):
        """ Load database files.
        """
        self.create_tables()
        self.parse_unicodedata('UnicodeData.txt')
        self.add_non_characters()
        self.parse_properties('DerivedAge.txt', 'age')
        self.parse_properties('Scripts.txt', 'script')
        self.parse_properties('HangulSyllableType.txt', 'hst')

        self.add_reserved_dicp()
        self.parse_property_present('DerivedCoreProperties.txt', 'dicp',
                                    'Default_Ignorable_Code_Point')
        self.parse_property_present('DerivedNormalizationProps.txt', 'fce',
                                    'Full_Composition_Exclusion')
        self.add_exceptions()
        self.assign_has_compat()
        self.assign_precis()
        self._conn.commit()

    def create_tables(self):
        """ Create database tables.
        """
        cur = self._conn.cursor()
        ddl = '''
            CREATE TABLE codepoints (
              cp         INTEGER PRIMARY KEY,
              name       TEXT NOT NULL,
              category   TEXT NOT NULL,
              combining  INTEGER NOT NULL,
              bidi       TEXT NOT NULL,
              decomp     TEXT NOT NULL,
              first_cp   INTEGER NOT NULL,  -- first cp in canonical decomp, -1 means no decomp, -2 means
              age        REAL,
              script     TEXT,
              hst        TEXT,       -- HangulSyllableType
              dicp       INT,        -- Default_Ignorable_Code_Point
              fce        INT,        -- Full_Composition_Exclusion
              has_compat INT,
              precis     TEXT
            )
        '''
        cur.execute(ddl)

    def parse_unicodedata(self, filename):
        """ Load data from UnicodeData.txt file.
        """
        cur = self._conn.cursor()
        for line in open(filename):
            cols = line.split(';')
            cp = int(cols[0], 16)
            name = cols[1]
            if name.endswith(', First>'):
                first = cp
            elif name.endswith(', Last>'):
                name = name[:-7] + ' %4.4x-%4.4x>' % (first, cp)
                for n in range(first, cp + 1):
                    self._insert(cur, n, name, cols[2], cols[3], cols[4],
                                 cols[5])
            else:
                self._insert(cur, cp, name, cols[2], cols[3], cols[4], cols[5])

    def add_non_characters(self):
        """ Add entries for non-characters.
        """
        cur = self._conn.cursor()
        for cp in range(0xfdd0, 0xfdef + 1):
            self._insert(cur, cp, '<noncharacter>', 'Cn', 0, '', '')
        for n in range(0, 17):
            cp1 = (n << 16) | 0xfffe
            cp2 = (n << 16) | 0xffff
            self._insert(cur, cp1, '<noncharacter>', 'Cn', 0, '', '')
            self._insert(cur, cp2, '<noncharacter>', 'Cn', 0, '', '')

    def add_reserved_dicp(self):
        """ Add entries for <reserved> chars that have the 'dicp' property.
        """
        cur = self._conn.cursor()
        self._insert(cur, 0x2065, '<reserved>', 'Cn', 0, '', '')
        for cp in range(0xfff0, 0xfff8 + 1):
            self._insert(cur, cp, '<reserved>', 'Cn', 0, '', '')
        self._insert(cur, 0xe0000, '<reserved>', 'Cn', 0, '', '')
        for cp in range(0xe0002, 0xe001f + 1):
            self._insert(cur, cp, '<reserved>', 'Cn', 0, '', '')
        for cp in range(0xe0080, 0xe00ff + 1):
            self._insert(cur, cp, '<reserved>', 'Cn', 0, '', '')
        for cp in range(0xe01f0, 0xe0fff + 1):
            self._insert(cur, cp, '<reserved>', 'Cn', 0, '', '')

    def parse_properties(self, filename, column):
        """ Load data from a Unicode property file.
        """
        cur = self._conn.cursor()
        for line in open(filename):
            line = line[:-1]
            if not line or line[0] == '#':
                continue
            m = self.property_regex.match(line)
            if not m:
                raise ValueError('Parse failed: %s' % line)
            self._set_column(cur, column, m.group(1), m.group(2), m.group(3))

    def parse_property_present(self, filename, column, value):
        """ Load data from a Unicode property file. Set `column` to 1 if we
        find `value`.
        """
        cur = self._conn.cursor()
        for line in open(filename):
            line = line[:-1]
            if not line or line[0] == '#':
                continue
            m = self.property_regex.match(line)
            if not m:
                raise ValueError('Parse failed: %s' % line)
            if m.group(3) == value:
                self._set_column(cur, column, m.group(1), m.group(2), 1)

    def _insert(self, cur, cp, name, category, combining, bidi, decomp):
        """ Insert a codepoint into table.
        """
        # Set `first_cp` depending on the value of the `decomp` field. If
        # `decomp` is empty, set first_cp to -1. If `decomp` is a compatibility
        # decomposition (starts with '<'), set first_cp to -2. Otherwise, set
        # first_cp to the character code of the first codepoint in `decomp`.
        if not decomp:
            first_cp = -1
        elif decomp.startswith('<'):
            first_cp = -2
        else:
            first_cp = int(decomp.split()[0], 16)
        cur.execute(
            'INSERT INTO codepoints (cp, name, category, combining, bidi, decomp, first_cp) VALUES (?, ?, ?, ?, ?, ?, ?)',
            (cp, name, category, int(combining), bidi, decomp, first_cp))

    def _set_column(self, cur, column, first, last, value):
        """ Set a specific column in the
        """
        first = int(first, 16)
        last = int(last, 16) if last else first
        sql = 'UPDATE codepoints SET %s=? WHERE cp=? AND %s IS NULL' % (column,
                                                                        column)
        for cp in range(first, last + 1):
            cur.execute(sql, (value, cp))
            if cur.rowcount != 1:
                print('failed update: %4.4x  %s=%s [%04x-%04x]' %
                      (cp, column, value, first, last))

    def add_exceptions(self):
        """ Add PRECIS exceptions.
        """
        sql = '''
        UPDATE codepoints SET precis = 'PVALID/exceptions' WHERE cp in (0x00DF,
            0x03C2, 0x06FD, 0x06FE, 0x0F0B, 0x3007);
        UPDATE codepoints SET precis = 'CONTEXTO/exceptions' WHERE cp in (0x00B7, 0x0375,
            0x05F3, 0x05F4, 0x30FB, 0x0660, 0x0661, 0x0662, 0x0663, 0x0664,
            0x0665, 0x0666, 0x0667, 0x0668, 0x0669, 0x06F0, 0x06F1, 0x06F2,
            0x06F3, 0x06F4, 0x06F5, 0x06F6, 0x06F7, 0x06F8, 0x06F9);
        UPDATE codepoints SET precis = 'DISALLOWED/exceptions' WHERE cp in (
            0x0640, 0x07FA, 0x302E, 0x302F, 0x3031, 0x3032, 0x3033, 0x3034,
            0x3035, 0x303B);
        '''
        cur = self._conn.cursor()
        cur.executescript(sql)

    def assign_precis(self):
        """ Assign precis derived property value to each codepoint.

        This is called after exceptions and backward_compatible have been
        assigned their precis properties.
        """
        sql = '''
            UPDATE codepoints SET precis = (
              CASE
              -- unassigned
              WHEN category = 'Cn' AND name != '<noncharacter>' THEN 'UNASSIGNED/unassigned'
              -- ascii7
              WHEN cp BETWEEN 0x21 AND 0x7E THEN 'PVALID/ascii7'
              -- join_control
              WHEN cp BETWEEN 0x200c AND 0x200d THEN 'CONTEXTJ/join_control'
              -- old_hangul_jamo
              WHEN hst IN ('L', 'V', 'T') THEN 'DISALLOWED/old_hangul_jamo'
              -- precis_ignorable_properties
              WHEN dicp == 1 OR name == '<noncharacter>' THEN 'DISALLOWED/precis_ignorable_properties'
              -- controls
              WHEN category = 'Cc' THEN 'DISALLOWED/controls'
              -- has_compat
              WHEN has_compat = 1 THEN 'FREE_PVAL/has_compat'
              -- letter_digits
              WHEN category IN ('Ll', 'Lu', 'Lo', 'Nd', 'Lm', 'Mn', 'Mc') THEN 'PVALID/letter_digits'
              -- other_letter_digits
              WHEN category IN ('Lt', 'Nl', 'No', 'Me') THEN 'FREE_PVAL/other_letter_digits'
              -- spaces
              WHEN category = 'Zs' THEN 'FREE_PVAL/spaces'
              -- symbols
              WHEN category IN ('Sm', 'Sc', 'Sk', 'So') THEN 'FREE_PVAL/symbols'
              -- punctuation
              WHEN category IN ('Pc', 'Pd', 'Ps', 'Pe', 'Pi', 'Pf', 'Po') THEN 'FREE_PVAL/punctuation'
              -- other
              ELSE 'DISALLOWED/other'
              END
            ) WHERE precis IS NULL
        '''
        cur = self._conn.cursor()
        cur.execute(sql)

    def assign_has_compat(self):
        """ Assign true to characters that have compatibility decompositions.
         For these, `normalize('NFKC', ch) != ch`.
        """
        cur = self._conn.cursor()
        # Set has_compat=1 for characters whose decomp field begins with '<' or
        # has a 'full composition exclusion' of 1.
        sql = '''
            UPDATE codepoints SET has_compat=1 WHERE decomp LIKE '<%' OR fce = 1
        '''
        cur.execute(sql)
        # The set of codepoints with compatibility decompositions is not complete
        # until we include the set of approximately 15 chars whose CANONICAL
        # decomposition has a further COMPATIBILITY decomposition.
        sql = '''
            UPDATE codepoints SET has_compat=1 WHERE cp IN (
                SELECT a.cp from codepoints a, codepoints b  WHERE b.cp = a.first_cp AND b.first_cp == -2
            )
        '''
        cur.execute(sql)

    def check_has_compat(self, ucd):
        """ Check that has_compat is set to 1 for every character where
        normalize(NFKC, ch) != ch.
        """
        cur = self._conn.cursor()
        sql = 'SELECT cp, has_compat, age FROM codepoints WHERE age <= %g' % ucd.version
        for cp, has_compat, age in cur.execute(sql):
            char = chr(cp)
            norm = ucd.normalize('NFKC', char)
            if has_compat == 1:
                if norm == char:
                    print('Invalid has_compat=1 for cp=%d, age=%s' % (cp, age))
            else:
                if norm != char:
                    print('Invalid has_compat=0 for cp=%d, age=%s' % (cp, age))

    def check_precis(self, ucd):
        """ Compare derived property computation to `precis` value in database.
        """
        from precis_i18n.derived import derived_property

        cur = self._conn.cursor()
        sql = 'SELECT cp, precis, age FROM codepoints WHERE age <= %g' % UCD.version
        for cp, precis, age in cur.execute(sql):
            prop = '%s/%s' % derived_property(cp, ucd)
            if prop != precis:
                print('Different precis value: %s vs %s for cp=%d, age=%s' %
                      (prop, precis, cp, age))


if __name__ == '__main__':
    try:
        db = UnicodeDatabase('unicode.db')
        db.load()
    except sqlite3.OperationalError:
        pass

    from precis_i18n.unicode import UnicodeData

    UCD = UnicodeData()
    db.check_has_compat(UCD)
    db.check_precis(UCD)



import unicodedata

from precis_i18n import get_profile

profile = get_profile('nicknamecasemapped:ToLower')


def _escape(s):
    return s.encode('unicode-escape').decode('ascii')


for cp in range(0x0110000):
    original = chr(cp)
    try:
        actual = profile.enforce(original)
        if actual != original:
            idempotent = profile.enforce(actual)
            if idempotent.strip() != actual.strip():
                print(_escape(original), unicodedata.name(original), ';',
                      unicodedata.decomposition(original))
    except UnicodeEncodeError:
        pass



import unittest

from precis_i18n.codepointset import CodepointSet


class TestCodepointSet(unittest.TestCase):
    def test_contains(self):
        cps = CodepointSet('0000\n')
        actual = [cp in cps for cp in range(-1, 4)]
        self.assertEqual(actual, [False, True, False, False, False])

        self.assertFalse(0x010FFFF in cps)

        cps = CodepointSet('0000..0001\n')
        actual = [cp in cps for cp in range(-1, 4)]
        self.assertEqual(actual, [False, True, True, False, False])

        cps = CodepointSet('0000\n0001\n0002')
        actual = [cp in cps for cp in range(-1, 4)]
        self.assertEqual(actual, [False, True, True, True, False])

        cps = CodepointSet('0000\n0002')
        actual = [cp in cps for cp in range(-1, 4)]
        self.assertEqual(actual, [False, True, False, True, False])

        cps = CodepointSet('10000..10FFFF')
        self.assertTrue(0x10FFFF in cps)
        self.assertFalse(0x110000 in cps)

    def test_len(self):
        cps = CodepointSet('0000\n')
        self.assertEqual(len(cps), 1)

        cps = CodepointSet('0000..0001\n')
        self.assertEqual(len(cps), 2)

        cps = CodepointSet('0000\n0001\n0002')
        self.assertEqual(len(cps), 3)

        cps = CodepointSet('0000\n0002')
        self.assertEqual(len(cps), 2)

        cps = CodepointSet('10000..10FFFF')
        self.assertEqual(len(cps), 0x10FFFF - 0x10000 + 1)

    def test_equals(self):
        cps = CodepointSet('0000..00FF')
        self.assertEqual(cps, CodepointSet('0000..00FF'))
        self.assertNotEqual(cps, CodepointSet('0000..00FE'))

        # Non-CodepointSet always not-equal.
        self.assertFalse(cps == 'what?')

    def test_repr(self):
        cps = CodepointSet('')
        self.assertEqual(repr(cps), "CodepointSet('')")

        cps = CodepointSet('0000')
        self.assertEqual(repr(cps), "CodepointSet('0000')")

        cps = CodepointSet('0000..00FF')
        self.assertEqual(repr(cps), "CodepointSet('0000..00FF')")

        cps = CodepointSet('0001..FFFF\n100000..10FFFF')
        self.assertEqual(repr(cps),
                         r"CodepointSet('0001..FFFF\n100000..10FFFF')")

        cps = CodepointSet('FFFF..1FFFF')
        self.assertEqual(repr(cps), "CodepointSet('FFFF..1FFFF')")

        cps = CodepointSet('10000..1FFFF')
        self.assertEqual(repr(cps), "CodepointSet('10000..1FFFF')")

        cps = CodepointSet('FFFE\n10000..1FFFF')
        self.assertEqual(repr(cps), r"CodepointSet('FFFE\n10000..1FFFF')")

    def test_coalesce(self):
        cps = CodepointSet('0000\n0001\n0002')
        self.assertEqual(cps, CodepointSet('0000..0002'))

        cps = CodepointSet('0000\n0002\n0003..0004')
        self.assertEqual(cps, CodepointSet('0000\n0002..0004'))

    def test_malformed_range(self):
        with self.assertRaises(ValueError):
            CodepointSet('0002..0000\n0001')

        with self.assertRaises(ValueError):
            CodepointSet('0000..0001\n0000..0001\n0002')

        with self.assertRaises(ValueError):
            CodepointSet('0000\n0002\n0002..0004')

        with self.assertRaises(ValueError):
            CodepointSet('110000')

        with self.assertRaises(ValueError):
            CodepointSet('0000\n000G')

    def test_even_odd(self):
        data = '\n'.join("%04X" % cp for cp in range(0, 10000, 2))
        cps = CodepointSet(data)
        for cp in range(10000):
            if cp in cps:
                self.assertTrue((cp % 2) == 0)
            else:
                self.assertFalse((cp % 2) == 0)

    def test_parse(self):
        cps = CodepointSet('A\nBB\n')
        self.assertEqual(repr(cps), r"CodepointSet('000A\n00BB')")

        cps = CodepointSet('AAA\nBBB..CCC\n')
        self.assertEqual(repr(cps), r"CodepointSet('0AAA\n0BBB..0CCC')")

        cps = CodepointSet('\n  \n # comment  \n   \n')
        self.assertEqual(repr(cps), "CodepointSet('')")



import unicodedata
from collections import Counter

import precis_i18n as precis


def _escape(s):
    return s.encode('unicode-escape').decode('ascii')


def _idempotent_ignoring_space(profile, value):
    result1 = profile.enforce(value)
    result2 = profile.enforce(result1)
    return result1.strip() == result2.strip()


results = Counter()
profile = precis.get_profile('NicknameCaseMapped:ToLower')

for cp in range(0x0110000):
    char = chr(cp)
    try:
        if not _idempotent_ignoring_space(profile, char):
            decomp = unicodedata.decomposition(char)
            kind = decomp.split()[0]
            if kind.startswith('<'):
                results[kind] += 1
            else:
                print(_escape(char), unicodedata.name(char))
    except UnicodeEncodeError:
        pass

print(results)



import json
import os
import unittest

import precis_i18n.codec  # noqa: F401

HERE = os.path.abspath(os.path.dirname(__file__))
GOLDEN_JSON = os.path.join(HERE, 'golden.json')

UCD_VERSION = precis_i18n.get_profile('FreeFormClass').ucd.version


class TestGolden(unittest.TestCase):
    def test_golden_json(self):
        with open(GOLDEN_JSON, encoding='ascii') as input_file:
            entries = json.load(input_file)

        for entry in entries:
            profile, input_, output, error = (entry['profile'], entry['input'],
                                              entry['output'], entry['error'])
            if 'unicode_version' in entry and UCD_VERSION < entry[
                    'unicode_version']:
                self.check_fails(profile, input_, output)
                continue

            if not error:
                self.check_allow(profile, input_, output)
            else:
                self.check_disallow(profile, input_, error)

    def check_allow(self, profile, input_, expected):
        # print('check_allow', profile, input_)
        try:
            actual = input_.encode(profile).decode('utf-8')
        except UnicodeEncodeError as ex:
            print('%s: %r' % (input_, ex))
            raise
        self.assertEqual(actual, expected)
        # Check that the profile encoding is idempotent. If the following
        # assertion fails, the profile is not idempotent.
        idempotent = actual.encode(profile).decode('utf-8')
        if idempotent != actual:
            print('\n"%s" not idempotent: "%s" => "%s" => "%s"' %
                  (profile, _escape(input_), _escape(actual),
                   _escape(idempotent)))
        self.assertEqual(idempotent, actual)

    def check_disallow(self, profile, input_, expected):
        # print('check_disallow', profile, input_)
        with self.assertRaisesRegex(UnicodeEncodeError, expected):
            input_.encode(profile)

    def check_fails(self, profile, input_, expected):
        """Check that output doesn't match or raises UnicodeEncodeError.

        This is used when the Unicode version is too low."""
        try:
            actual = input_.encode(profile).decode('utf-8')
        except UnicodeEncodeError:
            actual = None
        self.assertNotEqual(actual, expected)


def _escape(s):
    return s.encode('unicode_escape').decode('ascii')


if __name__ == '__main__':
    unittest.main(verbosity=2)



from typing import Generator, Tuple


class CodepointSet:

    def __init__(self, table: str) -> None:
        ...

    def __contains__(self, cp: int) -> bool:
        ...

    def __len__(self) -> int:
        ...

    def __eq__(self, rhs: object) -> bool:
        ...

    def __repr__(self) -> str:
        ...

    def items(self) -> Generator[Tuple[int, int], None, None]:
        ...



"""
Generate 'Kevin' table for README.
"""

import precis_i18n.codec  # noqa: F401

letter_k = [
    ord('K'), 0x212A, 0xFF2B, 0x039A, 0x1E32, 0x1E34, 0x2C69, 0xA740, 0xA742,
    0xA744, 0xA7A2, 0x24C0, 0x1F11A, 0x1F13A, 0x1F15A, 0x1F17A
]


def _escape(s):
    return s.encode('raw-unicode-escape').decode('ascii').replace(
        '\\', '&#92;')


def _xml_escape(s):
    return s.encode('ascii', errors='xmlcharrefreplace').decode('ascii')


def _column(s):
    try:
        # Leave ASCII strings alone.
        s.encode('ascii')
        return s
    except UnicodeEncodeError:
        pass
    return "%s (%s)" % (_xml_escape(s), _escape(s))


print(
    'Original String|UsernameCasePreserved|UsernameCaseMapped|NicknameCaseMapped'
)
print(
    '---------------|---------------------|------------------|------------------'
)

for k in letter_k:
    kevin = chr(k) + 'evin'

    try:
        case_preserved = kevin.encode('UsernameCasePreserved').decode('utf-8')
    except UnicodeEncodeError:
        case_preserved = 'DISALLOWED'

    try:
        case_mapped = kevin.encode('UsernameCaseMapped').decode('utf-8')
    except UnicodeEncodeError:
        case_mapped = 'DISALLOWED'

    nickname = kevin.encode('NicknameCaseMapped').decode('utf-8')

    col1 = _column(kevin)
    col2 = _column(case_preserved)
    col3 = _column(case_mapped)
    col4 = _column(nickname)

    print('%s | %s | %s | %s' % (col1, col2, col3, col4))



from precis_i18n.unicode import UnicodeData


def context_rule_error(value: str, offset: int, ucd: UnicodeData) -> str:
    ...

def rule_zero_width_nonjoiner(value: str, offset: int, ucd: UnicodeData) -> bool:
    ...

def rule_zero_width_joiner(value: str, offset: int, ucd: UnicodeData) -> bool:
    ...

def rule_middle_dot(value: str, offset: int, ucd: UnicodeData) -> bool:
    ...

def rule_greek_keraia(value: str, offset: int, ucd: UnicodeData) -> bool:
    ...

def rule_hebrew_punctuation(value: str, offset: int, ucd: UnicodeData) -> bool:
    ...

def rule_katakana_middle_dot(value: str, offset: int, ucd: UnicodeData) -> bool:
    ...

def rule_arabic_indic(value: str, offset: int, ucd: UnicodeData) -> bool:
    ...

def rule_extended_arabic_indic(value: str, offset: int, ucd: UnicodeData) -> bool:
    ...

_RULES = ...



import os
import re
import unittest

VERSIONS = [
    '6.1', '6.2', '6.3', '8.0', '9.0', '10.0', '11.0', '12.0', '12.1', '13.0',
    '14.0',
]

UNASSIGNED = 1

PROPS = {
    'UNASSIGNED': UNASSIGNED,
    'PVALID': 2,
    'FREE_PVAL': 3,
    'ID_DIS or FREE_PVAL': 3,  # found in IANA format only
    'DISALLOWED': 4,
    'CONTEXTJ': 5,
    'CONTEXTO': 6
}

LINE_REGEX = re.compile(
    r'^([0-9A-F]{4,6})-([0-9A-F]{4,6}) ([A-Z_]+)/[a-z0-9_]+$')

IANA_LINE_REGEX = re.compile(r'^([0-9A-F]{4,6})(-[0-9A-F]{4,6})?,([^,]+),.+$')

DIR_PATH = os.path.dirname(__file__)

# Allowed transitions between two versions V1 -> V2 for specific code points.
EXCEPTIONS = {
    0x111c9: (PROPS['FREE_PVAL'], PROPS['PVALID'])  # SHARADA SANDHI MARK
}


def _allowed_change(cp, tbl1, tbl2):
    """Return true if the transition is allowed.

    We allow changing from UNASSIGNED to anything. Also, in 10.0 -> 11.0, the
    code point 70089 (0x111c9, SHARADA SANDHI MARK) changed from FREE_PVAL to
    PVALID.
    """
    if tbl1 == UNASSIGNED:
        return True
    rule = EXCEPTIONS.get(cp)
    if rule:
        return (tbl1, tbl2) == rule
    return False


def _load_table(filename):
    """Load data from `derived-props-<version>.txt` file."""

    table = bytearray(0x110000)

    with open(filename) as fp:
        for line in fp:
            m = LINE_REGEX.match(line)
            assert m, 'Unexpected format: %s' % line

            lo, hi = int(m.group(1), 16), int(m.group(2), 16)
            prop = PROPS[m.group(3)]
            for cp in range(lo, hi + 1):
                table[cp] = prop

    # Check that all codepoints are assigned.
    for cp in range(0x110000):
        assert table[cp] != 0, 'Codepoint missing: %d' % cp

    return table


def _load_table_iana(filename):
    """Load table from IANA csv file."""

    table = bytearray(0x110000)

    with open(filename) as fp:
        for line in fp:
            # Ignore csv header.
            if line == 'Codepoint,Property,Description\n':
                continue

            m = IANA_LINE_REGEX.match(line)
            assert m, 'Unexpected format: %s' % line

            lo = int(m.group(1), 16)
            if m.group(2):
                hi = int(m.group(2)[1:], 16)
            else:
                hi = lo

            prop = PROPS[m.group(3)]
            for cp in range(lo, hi + 1):
                table[cp] = prop

    # Check that all codepoints are assigned.
    for cp in range(0x110000):
        assert table[cp] != 0, 'Codepoint missing: %d' % cp

    return table


def _load_tables():
    """Load data from all derived-props files."""

    tables = []
    file_template = os.path.join(DIR_PATH, 'derived-props-%s.txt')

    for version in VERSIONS:
        table = _load_table(file_template % version)
        tables.append((version, table))

    return tables


class TestDerivedPropsFiles(unittest.TestCase):
    def test_derived_props(self):
        """Check derived property values do not change as UCD version increases.

        The only allowed change is from UNASSIGNED to anything.
        """
        tables = _load_tables()

        for i in range(len(tables) - 1):
            # Compare table i to table i+1.
            # If table[i] != table[i+1] then table[i] must equal UNASSIGNED(1).
            ver1, tbl1 = tables[i]
            ver2, tbl2 = tables[i + 1]
            for j in range(0x110000):
                if not _allowed_change(j, tbl1[j], tbl2[j]):
                    self.assertEqual(tbl1[j], tbl2[j],
                                     'cp = %d (%s -> %s)' % (j, ver1, ver2))

    def test_iana_derived_props(self):
        """Compare IANA precis-tables to derived-props-6.3.txt"""

        iana_path = os.path.join(DIR_PATH, 'iana-precis-tables-6.3.0.csv')
        test_path = os.path.join(DIR_PATH, 'derived-props-6.3.txt')

        iana_table = _load_table_iana(iana_path)
        table = _load_table(test_path)

        # Tables should be identical.
        self.assertTrue(table == iana_table)



# Hunt for examples where order of NFC/Casefold makes a difference.

import unicodedata


def case(ch):
    return ch.casefold()


def nfc(ch):
    return unicodedata.normalize('NFC', ch)


def make(ch):
    return '%s%s' % (ch, unicodedata.normalize('NFD', ch))


for i in range(0, 0x010ffff):
    s = make(chr(i))
    if case(nfc(s)) != nfc(case(s)):
        print(s.encode('unicode-escape'))



"""Registers precis_i18n codec."""

import codecs

from precis_i18n import get_profile


def _make_encode(profile):
    def _encode(input, errors='strict'):
        if errors != 'strict':
            raise ValueError('invalid errors argument')
        return (profile.enforce(input).encode('utf-8'), len(input))

    return _encode


def _not_supported(input, errors='strict'):
    # pylint: disable=unused-argument
    raise NotImplementedError('decode not supported')


def search(name):
    """Search function registered for PRECIS codecs.

    Args:
        name (str): Codec name.

    Returns:
        CodecInfo: Encode/decode information or None if not found.
    """
    try:
        profile = get_profile(name)
    except KeyError:
        return None

    return codecs.CodecInfo(name=name,
                            encode=_make_encode(profile),
                            decode=_not_supported)


codecs.register(search)



from precis_i18n import get_profile
from precis_i18n.derived import derived_property
from precis_i18n.unicode import UnicodeData

UCD = UnicodeData()
profile = get_profile('UsernameCasePreserved')

count = 0
ascii = 0
for cp in range(0, 0x110000):
    _, reason = derived_property(cp, UCD)
    if reason == 'has_compat':
        try:
            result = profile.enforce(chr(cp)).encode('utf-8')
            print('%04x => %s' % (cp, result))

            if len(result) == 1 and result[0] < 128:
                ascii += 1
            count += 1
        except UnicodeEncodeError:
            pass
print('Exceptions: %d, %d ascii' % (count, ascii))



import unittest

from precis_i18n import get_profile
from precis_i18n.profile import Username
from precis_i18n.unicode import UnicodeData


class IdempotentTestCase(unittest.TestCase):
    def test_broken_profile(self):
        """Test that we can catch a profile that is not idempotent.
        """
        class _BrokenProfile(Username):
            def additional_mapping_rule(self, value):
                return '%s+' % value

        broken = _BrokenProfile(UnicodeData(), name='Broken')
        with self.assertRaisesRegex(ValueError, 'DISALLOWED/not_idempotent'):
            broken.enforce('x')

    def test_all_codepoints(self):
        """Verify all individual code points are idempotent.
        """
        profiles = [
            get_profile(profile)
            for profile in ('UsernameCaseMapped', 'NicknameCaseMapped')
        ]
        for cp in range(0x0110000):
            original = chr(cp)
            for profile in profiles:
                try:
                    profile.enforce(original)
                except UnicodeEncodeError as ex:
                    if 'not_idempotent' in str(ex):
                        raise



import codecs
import unittest

import precis_i18n.codec


class TestCodec(unittest.TestCase):
    def test_encode(self):
        self.assertEqual('Juliet'.encode('UsernameCasePreserved'), b'Juliet')
        self.assertEqual('Juliet'.encode('UsernameCaseMapped'), b'juliet')
        self.assertEqual(' pass \u1FBF\u3000'.encode('OpaqueString'),
                         b' pass \xe1\xbe\xbf ')
        self.assertEqual(' Juliet '.encode('NicknameCaseMapped'), b'juliet')
        self.assertEqual('Juliet'.encode('IdentifierClass'), b'Juliet')
        self.assertEqual('Juliet'.encode('FreeFormClass'), b'Juliet')
        self.assertEqual('Juliet'.encode('UsernameCaseMapped:ToLower'),
                         b'juliet')
        self.assertEqual('Juliet'.encode('UsernameCaseMapped_ToLower'),
                         b'juliet')

    def test_decode(self):
        with self.assertRaises(NotImplementedError):
            b'Juliet'.decode('UsernameCasePreserved')

    def test_encode_errors(self):
        # errors must be 'strict'; 'replace' and others are not supported.
        with self.assertRaises(ValueError):
            'Juliet'.encode('opaquestring', errors='replace')
        # non-matching codec names shouldn't work.
        with self.assertRaises(LookupError) as cm:
            'Juliet'.encode('opaquestring_nonexistant')
        # Exception must be LookupError (not KeyError or IndexError).
        self.assertIs(type(cm.exception), LookupError)

    def test_search_function(self):
        _search = precis_i18n.codec.search

        # Check search function result.
        codec_info = _search('usernamecasepreserved')
        self.assertIsInstance(codec_info, codecs.CodecInfo)

        # Search function must return None for non-existant codec.
        codec_info = _search('opaquestring_nonexistant')
        self.assertIs(codec_info, None)



# test_precis.py

import platform
import sys
import unittest

import precis_i18n.context as pc
from precis_i18n.baseclass import FreeFormClass, IdentifierClass
from precis_i18n.bidi import bidi_rule, has_rtl
from precis_i18n.derived import derived_property
from precis_i18n.unicode import UnicodeData, _version_to_float

_PYPY = (platform.python_implementation() == 'PyPy')
_PY3_11 = sys.version_info[:2] >= (3, 11)

UCD = UnicodeData()

# Example characters for bidirectional properties. Used in testing Bidi Rule.
L = 'A'
R = '\u05d0'
AL = '\u0621'
EN = '0'
AN = '\U00010e60'
NSM = '\u0300'
P = '*'


class TestBidiRule(unittest.TestCase):
    def test_bidi_rule_ltr(self):
        self.assertTrue(bidi_rule(L, UCD))
        self.assertTrue(bidi_rule(L + P + L, UCD))
        self.assertFalse(bidi_rule(P + L, UCD))
        self.assertFalse(bidi_rule(L + P, UCD))
        self.assertFalse(bidi_rule(L + R + L, UCD))

    def test_bidi_rule_rtl(self):
        self.assertTrue(bidi_rule(R, UCD))
        self.assertTrue(bidi_rule(R + R, UCD))
        self.assertTrue(bidi_rule(AL + R, UCD))
        self.assertTrue(bidi_rule(AL + EN + R, UCD))
        self.assertTrue(bidi_rule(AL + R + EN, UCD))
        self.assertTrue(bidi_rule(R + P + R, UCD))
        self.assertTrue(bidi_rule(R + P + R + NSM + NSM, UCD))
        self.assertTrue(bidi_rule(R + AN + NSM + AN + R, UCD))
        self.assertFalse(bidi_rule(P + R, UCD))
        self.assertFalse(bidi_rule(R + P, UCD))
        self.assertFalse(bidi_rule(R + L + R, UCD))
        self.assertFalse(bidi_rule(EN + R, UCD))
        self.assertFalse(bidi_rule(R + AN + EN, UCD))
        self.assertFalse(bidi_rule(R + EN + AN + R, UCD))

    def test_has_rtl(self):
        self.assertFalse(has_rtl('Juliet+', UCD))
        self.assertTrue(has_rtl('\u05d0+', UCD))


class TestPrecisIdentifierClass(unittest.TestCase):
    def test_valid_identifier(self):
        ident = IdentifierClass(UCD)
        self.assertEqual(ident.name, 'IdentifierClass')
        self.assertEqual(ident.enforce('abc'), 'abc')
        self.assertEqual(ident.enforce('123'), '123')
        self.assertEqual(ident.enforce('\u0660\u0661\u0662\u0669'),
                         '\u0660\u0661\u0662\u0669')
        self.assertEqual(ident.enforce('\u0370\u0371'), '\u0370\u0371')
        # CONTEXTJ
        self.assertEqual(ident.enforce('\u094d\u200c'), '\u094d\u200c')

    def test_invalid_identifier(self):
        ident = IdentifierClass(UCD)
        # DISALLOWED/spaces
        with self.assertRaisesRegex(
                UnicodeEncodeError,
                r"'IdentifierClass' codec can't encode character '\\x20' in position 0: DISALLOWED/spaces"
        ):
            ident.enforce(' ')

        # DISALLOWED/precis_ignorable_properties
        with self.assertRaisesRegex(
                UnicodeEncodeError,
                r"'IdentifierClass' codec can't encode character '\\xad' in position 0: DISALLOWED/precis_ignorable_properties"
        ):
            ident.enforce('\xad')

        # DISALLOWED/old_hangul_jamo
        with self.assertRaisesRegex(
                UnicodeEncodeError,
                r"'IdentifierClass' codec can't encode character '\\u1100' in position 0: DISALLOWED/old_hangul_jamo"
        ):
            ident.enforce('\u1100')

        # DISALLOWED/has_compat
        with self.assertRaisesRegex(
                UnicodeEncodeError,
                r"'IdentifierClass' codec can't encode character '\\u1fbf' in position 0: DISALLOWED/has_compat"
        ):
            ident.enforce('\u1FBF')


class TestPrecisFreeformClass(unittest.TestCase):
    def test_valid_freeform(self):
        free = FreeFormClass(UCD)
        self.assertEqual(free.name, 'FreeFormClass')

        self.assertEqual(free.enforce('abc'), 'abc')
        self.assertEqual(free.enforce('123'), '123')
        self.assertEqual(free.enforce('\u0660\u0661\u0662\u0669'),
                         '\u0660\u0661\u0662\u0669')
        self.assertEqual(free.enforce('\u0370\u0371'), '\u0370\u0371')
        self.assertEqual(free.enforce(' '), ' ')
        self.assertEqual(free.enforce('\u1FBF'), '\u1FBF')

    def test_invalid_freeform(self):
        free = FreeFormClass(UCD)

        # DISALLOWED/precis_ignorable_properties
        with self.assertRaisesRegex(
                UnicodeEncodeError,
                r"'FreeFormClass' codec can't encode character '\\xad' in position 0: DISALLOWED/precis_ignorable_properties"
        ):
            free.enforce('\xad')

        # DISALLOWED/old_hangul_jamo
        with self.assertRaisesRegex(
                UnicodeEncodeError,
                r"'FreeFormClass' codec can't encode character '\\u1100' in position 0: DISALLOWED/old_hangul_jamo"
        ):
            free.enforce('\u1100')

        # DISALLOWED/other (surrogates)
        with self.assertRaisesRegex(
                UnicodeEncodeError,
                r"'FreeFormClass' codec can't encode character '\\ud800' in position 0: DISALLOWED/other"
        ):
            if _PYPY:
                # pypy3-v5.5.0 treats surrogate pairs in .pyc files differently. (issue #2441)
                free.enforce('\ud800 \udc00')
            else:
                free.enforce('\ud800\udc00')


class TestDerivedProperty(unittest.TestCase):
    def test_derived_property(self):
        self.assertEqual(derived_property(0, UCD), ('DISALLOWED', 'controls'))
        self.assertEqual(derived_property(0x10FFFF, UCD),
                         ('DISALLOWED', 'precis_ignorable_properties'))
        self.assertEqual(derived_property(0x30, UCD), ('PVALID', 'ascii7'))
        self.assertEqual(derived_property(0x20, UCD), ('FREE_PVAL', 'spaces'))
        self.assertEqual(derived_property(0xad, UCD),
                         ('DISALLOWED', 'precis_ignorable_properties'))

        with self.assertRaises(ValueError):
            derived_property(0x110000, UCD)

        # In Unicode 9.0, U+08E2 is 'DISALLOWED'. Before Unicode 9.0, it is
        # UNASSIGNED.
        prop = derived_property(0x08e2, UCD)[0]
        if UCD.version >= 9.0:
            self.assertEqual(prop, 'DISALLOWED')
        else:
            self.assertEqual(prop, 'UNASSIGNED')


class TestPrecisContextRule(unittest.TestCase):
    def test_rule_zero_width_nonjoiner(self):
        # We're going to use a872 and 0622 in some tests. Make sure they aren't
        # combining_virama().
        self.assertFalse(UCD.combining_virama(0xa872))
        self.assertFalse(UCD.combining_virama(0x0622))
        # Valid: combining_virama before
        self.assertTrue(pc.rule_zero_width_nonjoiner('\u094d\u200c', 1, UCD))
        # Invalid: invalid join_type
        self.assertFalse(pc.rule_zero_width_nonjoiner('\ua872\u200c', 1, UCD))
        # Invalid: undefined before
        with self.assertRaises(IndexError):
            pc.rule_zero_width_nonjoiner('\u200c', 0, UCD)
        # Valid: jointype(L J R)
        self.assertTrue(
            pc.rule_zero_width_nonjoiner('\ua872\u200c\u0622', 1, UCD))
        # Invalid: jointype(R J L)
        self.assertFalse(
            pc.rule_zero_width_nonjoiner('\u0622\u200c\ua872', 1, UCD))

    def test_rule_zero_width_joiner(self):
        # Valid: combining_virama before
        self.assertTrue(pc.rule_zero_width_joiner('\u094d\u200d', 1, UCD))
        # Invalid: no combining_virama before
        self.assertFalse(pc.rule_zero_width_joiner('A\u200d', 1, UCD))
        # Invalid: no combining_virama before, jointype(L J R)
        self.assertFalse(
            pc.rule_zero_width_joiner('\ua872\u200d\u0622', 1, UCD))

    def test_rule_middle_dot(self):
        # Valid: 6c b7 6c
        self.assertTrue(pc.rule_middle_dot('\u006c\u00b7\u006c', 1, UCD))
        # Invalid before: 6d b7 6c
        self.assertFalse(pc.rule_middle_dot('\u006d\u00b7\u006c', 1, UCD))
        # Invalid after: 6c b7 6d
        self.assertFalse(pc.rule_middle_dot('\u006c\u00b7\u006d', 1, UCD))
        # Invalid: undefined before
        with self.assertRaises(IndexError):
            pc.rule_middle_dot('\u00b7\u006c', 0, UCD)
        # Invalid: undefined after
        with self.assertRaises(IndexError):
            pc.rule_middle_dot('\u006c\u00b7', 1, UCD)

    def test_rule_greek_keraia(self):
        # Valid: 0375 03ff
        self.assertTrue(pc.rule_greek_keraia('\u0375\u03ff', 0, UCD))
        # Invalid: 0375 1d25
        self.assertFalse(pc.rule_greek_keraia('\u0375\u1d25', 0, UCD))
        # Invalid: undefined after
        with self.assertRaises(IndexError):
            pc.rule_greek_keraia('\u0375', 0, UCD)

    def test_rule_hebrew_punctuation(self):
        # Valid: 0591 05f3
        self.assertTrue(pc.rule_hebrew_punctuation('\u0591\u05f3', 1, UCD))
        # Valid: 0591 05f4
        self.assertTrue(pc.rule_hebrew_punctuation('\u0591\u05f4', 1, UCD))
        # Invalid: 0031 05f3
        self.assertFalse(pc.rule_hebrew_punctuation('\u0031\u05f3', 1, UCD))
        # Invalid: 0031 05f4
        self.assertFalse(pc.rule_hebrew_punctuation('\u0031\u05f4', 1, UCD))
        # Invalid: undefined after
        with self.assertRaises(IndexError):
            pc.rule_hebrew_punctuation('\u05f3', 0, UCD)

    def test_katatana_middle_dot(self):
        # Valid: 0x30fb 0x2e99
        self.assertTrue(pc.rule_katakana_middle_dot('\u30fb\u2e99', 0, UCD))
        # Valid: 0x30f0 0x30fb 0x0021
        self.assertTrue(
            pc.rule_katakana_middle_dot('\u30f0\u30fb\u0021', 1, UCD))
        # Invalid: 0x30fb 0x3006
        self.assertFalse(pc.rule_katakana_middle_dot('\u30fb\u3006', 0, UCD))
        # Invalid: 0x30fb 0x0021
        self.assertFalse(pc.rule_katakana_middle_dot('\u0021\u30fb', 1, UCD))
        # Invalid: 0x30fb
        self.assertFalse(pc.rule_katakana_middle_dot('\u30fb', 0, UCD))

    def test_arabic_indic(self):
        # Valid: 0x0660 0x0661 0x0662 0x0669
        self.assertTrue(
            pc.rule_arabic_indic('\u0660\u0661\u0662\u0669', 0, UCD))
        # Invalid: 0x660 0x0661 0x0662 0x06f0
        self.assertFalse(
            pc.rule_arabic_indic('\u0660\u0661\u0662\u06f0', 0, UCD))

    def test_extended_arabic_indic(self):
        # Valid: 0x06f0 0x06f1 0x06f2 0x06f9
        self.assertTrue(
            pc.rule_extended_arabic_indic('\u06f0\u06f1\u06f2\u06f9', 0, UCD))
        # Invalid: 0x6f0 0x06f1 0x06f2 0x0660
        self.assertFalse(
            pc.rule_extended_arabic_indic('\u06f0\u06f1\u06f2\u0660', 0, UCD))

    def test_context_rule(self):
        def _context_rule(value, offset, ucd):
            return not pc.context_rule_error(value, offset, ucd)

        # 1. rule_zero_width_nonjoiner
        # Valid: combining_virama before
        self.assertTrue(_context_rule('\u094d\u200c', 1, UCD))
        # Invalid: invalid join_type
        self.assertFalse(_context_rule('\ua872\u200c', 1, UCD))
        # Invalid: undefined before
        self.assertFalse(_context_rule('\u200c', 0, UCD))
        # Valid: jointype(L J R)
        self.assertTrue(_context_rule('\ua872\u200c\u0622', 1, UCD))
        # Invalid: jointype(R J L)
        self.assertFalse(_context_rule('\u0622\u200c\ua872', 1, UCD))

        # 2. rule_zero_width_joiner
        # Valid: combining_virama before
        self.assertTrue(_context_rule('\u094d\u200d', 1, UCD))
        # Invalid: no combining_virama before
        self.assertFalse(_context_rule('A\u200d', 1, UCD))
        # Invalid: no combining_virama before, jointype(L J R)
        self.assertFalse(_context_rule('\ua872\u200d\u0622', 1, UCD))

        # 3. rule_middle_dot
        # Valid: 6c b7 6c
        self.assertTrue(_context_rule('\u006c\u00b7\u006c', 1, UCD))
        # Invalid before: 6d b7 6c
        self.assertFalse(_context_rule('\u006d\u00b7\u006c', 1, UCD))
        # Invalid after: 6c b7 6d
        self.assertFalse(_context_rule('\u006c\u00b7\u006d', 1, UCD))
        # Invalid: undefined before
        self.assertFalse(_context_rule('\u00b7\u006c', 0, UCD))
        # Invalid: undefined after
        self.assertFalse(_context_rule('\u006c\u00b7', 1, UCD))

        # 4. rule_greek_keraia
        # Valid: 0375 03ff
        self.assertTrue(_context_rule('\u0375\u03ff', 0, UCD))
        # Invalid: 0375 1d25
        self.assertFalse(_context_rule('\u0375\u1d25', 0, UCD))
        # Invalid: undefined after
        self.assertFalse(_context_rule('\u0375', 0, UCD))

        # 5. rule_hebrew_punctuation
        # Valid: 0591 05f3
        self.assertTrue(_context_rule('\u0591\u05f3', 1, UCD))
        # Valid: 0591 05f4
        self.assertTrue(_context_rule('\u0591\u05f4', 1, UCD))
        # Invalid: 0031 05f3
        self.assertFalse(_context_rule('\u0031\u05f3', 1, UCD))
        # Invalid: 0031 05f4
        self.assertFalse(_context_rule('\u0031\u05f4', 1, UCD))
        # Invalid: undefined after
        self.assertFalse(_context_rule('\u05f3', 0, UCD))
        # Valid: 05EF 05f3 (Unicode >= 11.0)
        self.assertTrue(_context_rule('\u05EF\u05f3', 1, UCD))

        # 6. katakana_middle_dot
        # Valid: 0x30fb 0x2e99
        self.assertTrue(_context_rule('\u30fb\u2e99', 0, UCD))
        # Valid: 0x30f0 0x30fb 0x0021
        self.assertTrue(_context_rule('\u30f0\u30fb\u0021', 1, UCD))
        # Invalid: 0x30fb 0x3006
        self.assertFalse(_context_rule('\u30fb\u3006', 0, UCD))
        # Invalid: 0x30fb 0x0021
        self.assertFalse(_context_rule('\u0021\u30fb', 1, UCD))
        # Invalid: 0x30fb
        self.assertFalse(_context_rule('\u30fb', 0, UCD))
        # Valid: 0x30fb 0x3400 (Unicode >= X)
        self.assertTrue(_context_rule('\u30fb\u3400', 0, UCD))

        # 7. arabic_indic
        # Valid: 0x0660 0x0661 0x0662 0x0669
        self.assertTrue(_context_rule('\u0660\u0661\u0662\u0669', 0, UCD))
        # Invalid: 0x660 0x0661 0x0662 0x06f0
        self.assertFalse(_context_rule('\u0660\u0661\u0662\u06f0', 0, UCD))

        # 8. extended_arabic_indic
        # Valid: 0x06f0 0x06f1 0x06f2 0x06f9
        self.assertTrue(_context_rule('\u06f0\u06f1\u06f2\u06f9', 0, UCD))
        # Invalid: 0x6f0 0x06f1 0x06f2 0x0660
        self.assertFalse(_context_rule('\u06f0\u06f1\u06f2\u0660', 0, UCD))

        # 9. No rule matches.
        with self.assertRaises(KeyError):
            _context_rule('a', 0, UCD)


class TestPrecisUnicodeData(unittest.TestCase):
    def test_width_map(self):
        self.assertEqual(UCD.width_map('\uff00\uff01\uff02\uffe3\uffef'),
                         '\uff00!"\uffe3\uffef')

    def test_replace_whitespace(self):
        self.assertEqual(
            UCD.map_nonascii_space_to_ascii(
                ' .\u00a0.\u1680 .\u2000.\u200A.\u202F.\u205F.\u3000'),
            ' . .  . . . . . ')

    def test_default_ignorable_code_point(self):
        self.assertTrue(UCD.default_ignorable(0x00ad))
        self.assertFalse(UCD.default_ignorable(0x00ae))

    def test_has_compat(self):
        self.assertFalse(UCD.has_compat(0x31))
        self.assertTrue(UCD.has_compat(0xff01))
        self.assertTrue(UCD.has_compat(0x212B))

    def test_control(self):
        self.assertFalse(UCD.control(0x20))
        self.assertTrue(UCD.control(0x0A))
        self.assertTrue(UCD.control(0x80))

    def test_noncharacter(self):
        self.assertTrue(UCD.noncharacter(0x02FFFF))
        self.assertTrue(UCD.noncharacter(0xfdef))
        self.assertFalse(UCD.noncharacter(0xfdf0))

    def test_old_hangul_jamo(self):
        self.assertTrue(UCD.old_hangul_jamo(0xa97c))
        self.assertFalse(UCD.old_hangul_jamo(0xa97d))

    def test_greek_script(self):
        self.assertTrue(UCD.greek_script(0x0373))
        self.assertFalse(UCD.greek_script(0x0374))

    def test_hebrew_script(self):
        self.assertTrue(UCD.hebrew_script(0x05c7))
        self.assertFalse(UCD.hebrew_script(0x05c8))

    def test_hiragana_katakana_han_script(self):
        self.assertTrue(UCD.hiragana_katakana_han_script(0x1f200))
        self.assertFalse(UCD.hiragana_katakana_han_script(0x1f201))
        self.assertTrue(UCD.hiragana_katakana_han_script(0xff9d))
        self.assertFalse(UCD.hiragana_katakana_han_script(0xff9e))
        self.assertTrue(UCD.hiragana_katakana_han_script(0x2fd5))
        self.assertFalse(UCD.hiragana_katakana_han_script(0x2fd6))
        self.assertFalse(UCD.hiragana_katakana_han_script(0x3006))
        self.assertFalse(UCD.hiragana_katakana_han_script(0x30fb))

    def test_combining_virama(self):
        self.assertTrue(UCD.combining_virama(0x1714))
        if _PY3_11:
            self.assertTrue(UCD.combining_virama(0x1715))
        else:
            self.assertFalse(UCD.combining_virama(0x1715))
        self.assertFalse(UCD.combining_virama(0x1716))

    def test_arabic_indic(self):
        self.assertTrue(UCD.arabic_indic(0x669))
        self.assertFalse(UCD.arabic_indic(0x66A))

    def test_extended_arabic_indic(self):
        self.assertTrue(UCD.extended_arabic_indic(0x06f9))
        self.assertFalse(UCD.extended_arabic_indic(0x06fa))

    def test_valid_join_type(self):
        # Valid: L J R
        self.assertTrue(UCD.valid_jointype('\ua872\u200c\u0622', 1))
        # Valid: L T J T R
        self.assertTrue(UCD.valid_jointype('\ua872\u00ad\u200c\u00ad\u0622',
                                           2))
        # Valid: D J D
        self.assertTrue(UCD.valid_jointype('\u0620\u200c\u0620', 1))
        # Valid: D T J T D
        self.assertTrue(UCD.valid_jointype('\u0626\u0300\u200c\u0301\u0628',
                                           2))

        # Invalid: R J L
        self.assertFalse(UCD.valid_jointype('\u0622\u200c\ua872', 1))
        # Invalid: R T J T L
        self.assertFalse(
            UCD.valid_jointype('\u0622\u00ad\u200c\u00ad\ua872', 2))
        # Invalid: J
        self.assertFalse(UCD.valid_jointype('\u200c', 0))
        # Invalid: T J T
        self.assertFalse(UCD.valid_jointype('\u00ad\u200c\u00ad', 1))
        # Invalid: U T J T U
        self.assertFalse(
            UCD.valid_jointype('\u0031\u0300\u200c\u0301\u0032', 2))

        # Valid: L J R   (Unicode >= 14.0)
        self.assertTrue(UCD.valid_jointype('\U00010D00\u200c\u088E', 1))
        # Valid: L T J T R  (Unicode >= 14.0)
        self.assertTrue(UCD.valid_jointype('\U00010D00\u07fd\u200c\u07fd\u088E', 2))
        # Valid: D J D  (Unicode >= 14.0)
        self.assertTrue(UCD.valid_jointype('\u0886\u200c\u0886', 1))


    def test_version_to_float(self):
        self.assertEqual(_version_to_float('8.0.0'), 8.0)
        self.assertEqual(_version_to_float('6.3.1'), 6.3)
        with self.assertRaises(ValueError):
            _version_to_float('8.0')


if __name__ == '__main__':
    unittest.main(verbosity=2)



from typing import Optional, Union

from precis_i18n.baseclass import BaseClass
from precis_i18n.unicode import UnicodeData


class Profile:
    def __init__(self, base: BaseClass, name: str, casemap: Optional[str] = ...) -> None:
        ...
    
    @property
    def base(self) -> BaseClass:
        ...
    
    @property
    def name(self) -> str:
        ...
    
    def enforce(self, value: Union[bytes, str]) -> str:
        ...
    
    def apply_five_rules(self, value: str) -> str:
        ...
    
    def width_mapping_rule(self, value: str) -> str:
        ...
    
    def additional_mapping_rule(self, value: str) -> str:
        ...
    
    def case_mapping_rule(self, value: str) -> str:
        ...
    
    def normalization_rule(self, value: str) -> str:
        ...
    
    def directionality_rule(self, value: str) -> str:
        ...
    
    def idempotence_check(self, value: str) -> str:
        ...
    


class Username(Profile):
    def __init__(self, ucd: UnicodeData, name: str, casemap: Optional[str] = ...) -> None:
        ...
    
    def width_mapping_rule(self, value: str) -> str:
        ...
    
    def directionality_rule(self, value: str) -> str:
        ...
    


class OpaqueString(Profile):
    def __init__(self, ucd: UnicodeData, name: str) -> None:
        ...
    
    def additional_mapping_rule(self, value: str) -> str:
        ...
    


class Nickname(Profile):
    def __init__(self, ucd: UnicodeData, name: str, casemap: Optional[str] = ...) -> None:
        ...
    
    def additional_mapping_rule(self, value: str) -> str:
        ...
    
    def normalization_rule(self, value: str) -> str:
        ...
    
    def idempotence_check(self, value: str) -> str:
        ...
    





"""A setuptools based setup module for precis_i18n.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

import re
from os import path

# Always prefer setuptools over distutils
from setuptools import setup

here = path.abspath(path.dirname(__file__))
description_path = path.join(here, 'README.rst')
version_path = path.join(here, 'precis_i18n', '__init__.py')

# Read README.rst.
with open(description_path, encoding='utf-8') as f:
    long_description = f.read()

# Extract version number.
with open(version_path, encoding='utf-8') as f:
    version_regex = re.compile(r"(?m)__version__\s*=\s*'(\d+\.\d+\.\d+)'")
    version = version_regex.search(f.read()).group(1)


# Running `python setup.py test` should run unit tests (see `test_suite`).
def all_tests():
    import unittest
    test_loader = unittest.TestLoader()
    return test_loader.discover('test')


setup(
    name='precis_i18n',
    packages=['precis_i18n'],
    version=version,
    license='MIT',
    description='PRECIS-i18n: Internationalized Usernames and Passwords',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    keywords='precis codec username password',

    # The project's main homepage and author.
    url='https://github.com/byllyfish/precis_i18n',
    author='William W. Fisher',
    author_email='william.w.fisher@gmail.com',

    package_data={
        'precis_i18n': ['py.typed', '*.pyi'],
    },

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development :: Internationalization'
    ],
    zip_safe=True,
    test_suite="setup.all_tests")



import unittest

import precis_i18n
from precis_i18n import get_profile
from precis_i18n.unicode import UnicodeData


class TestGetProfile(unittest.TestCase):
    def test_missing(self):
        with self.assertRaises(KeyError):
            get_profile('_does_not_exist_')

    def test_unicodedata_arg(self):
        import unicodedata
        profile = get_profile('UsernameCasePreserved', unicodedata=unicodedata)
        self.assertEqual(profile.enforce('E\u0301\u0301\u0301'),
                         '\u00c9\u0301\u0301')


class TestUsernameCasePreserved(unittest.TestCase):
    def test_enforce(self):
        profile = get_profile('UsernameCasePreserved')
        self.assertEqual(profile.enforce('Juliet'), 'Juliet')
        self.assertEqual(profile.enforce('J*'), 'J*')
        self.assertEqual(profile.enforce('E\u0301\u0301\u0301'),
                         '\u00c9\u0301\u0301')
        self.assertEqual(profile.enforce(b'Juliet'), 'Juliet')

        self.profile_fail(profile, '', 'empty')
        self.profile_fail(profile, ' J', 'spaces')
        self.profile_fail(profile, '\u05d0*', 'bidi_rule')

    def profile_fail(self, profile, value, reason):
        with self.assertRaisesRegex(UnicodeEncodeError, reason):
            profile.enforce(value)

    def test_invalid_argument(self):
        profile = get_profile('UsernameCasePreserved')
        with self.assertRaisesRegex(ValueError, 'not a string'):
            profile.enforce(1)

    def test_identifier_oddities(self):
        # Make a list of all codepoints < 10,000 which are allowed in the
        # UsernameCasePreserved profile even though they are not allowed in
        # IdentifierClass.
        profile = get_profile('UsernameCasePreserved')
        allowed = []
        for cp in range(0, 10000):
            try:
                profile.enforce(chr(cp))
                try:
                    profile.base.enforce(chr(cp))
                except UnicodeEncodeError:
                    allowed.append(cp)
            except UnicodeEncodeError:
                pass
        self.assertEqual(allowed, [
            832, 833, 835, 836, 884, 894, 2392, 2393, 2394, 2395, 2396, 2397,
            2398, 2399, 2524, 2525, 2527, 2611, 2614, 2649, 2650, 2651, 2654,
            2908, 2909, 3907, 3917, 3922, 3927, 3932, 3945, 3955, 3957, 3958,
            3960, 3969, 3987, 3997, 4002, 4007, 4012, 4025, 8049, 8051, 8053,
            8055, 8057, 8059, 8061, 8123, 8126, 8137, 8139, 8147, 8155, 8163,
            8171, 8175, 8185, 8187, 8486, 8490, 8491
        ])


class TestUsernameCaseMapped(unittest.TestCase):
    def test_enforce(self):
        profile = get_profile('UsernameCaseMapped')
        self.assertEqual(profile.enforce('Juliet'), 'juliet')
        self.assertEqual(profile.enforce('E\u0301\u0301\u0301'),
                         '\u00e9\u0301\u0301')


class TestNickname(unittest.TestCase):
    def test_enforce(self):
        profile = get_profile('Nickname')
        self.assertEqual(profile.enforce('Juliet'), 'juliet')
        self.assertEqual(profile.enforce('E\u0301\u0301\u0301'),
                         '\u00e9\u0301\u0301')
        self.assertEqual(profile.enforce('\u03d4'), '\u03cb')


class TestNicknameCaseMapped(unittest.TestCase):
    def test_enforce(self):
        profile = get_profile('NicknameCaseMapped:ToLower')
        self.assertEqual(profile.enforce('Juliet'), 'juliet')
        self.assertEqual(profile.enforce('E\u0301\u0301\u0301'),
                         '\u00e9\u0301\u0301')
        self.assertEqual(profile.enforce('\u03d4'), '\u03cb')


class TestUsername(unittest.TestCase):
    def test_constructor(self):
        profile = precis_i18n.profile.Username(UnicodeData(), 'name', 'lower')
        self.assertEqual(profile.enforce('Fu\u00dfball'), 'fu\u00dfball')

    def test_constructor_fail(self):
        with self.assertRaises(ValueError):
            precis_i18n.profile.Username(UnicodeData(), 'name',
                                         'unsupported-arg')



#!/bin/env python3
#
# Take a source file as input and produce a json file.
#
#  make_golden.py < golden_source.txt > golden.json
#

import json
import re
import sys
from collections import OrderedDict

import precis_i18n.codec  # noqa: F401

PROFILES = [
    'UsernameCasePreserved', 'UsernameCaseMapped', 'OpaqueString',
    'NicknameCaseMapped', 'UsernameCaseMapped:ToLower',
    'NicknameCasePreserved', 'NicknameCaseMapped:ToLower', 'FreeFormClass',
    'IdentifierClass', 'UsernameCaseMapped:CaseFold',
    'NicknameCaseMapped:CaseFold'
]

_ANY_PROFILE = re.compile(r'.')

EXCEPTIONS = {
    # ToLower difference before Unicode 8.0. The lower case characters weren't
    # added until Unicode 8.0.
    '\u13da\u13a2\u13b5\u13ac\u13a2\u13ac\u13d2':
    [(re.compile(r'.+CaseMapped(:ToLower)?$'), 8.0)],
    # U+1AB6 was introduced in 7.0.
    '\u05d0\u1ab6\u05d1': [(_ANY_PROFILE, 7.0)],
    # U+0111C9 changed to PVALID in 11.0. It was introduced in 8.0.
    '\U000111c9': [(re.compile(r'^(?:Username.*|IdentifierClass)$'), 11.0),
                   (_ANY_PROFILE, 8.0)],

    # U+05EF was introduced in 11.0.
    '\u05EF\u05f3': [(_ANY_PROFILE, 11.0)],

    # U+088E, U+0886 were introduced in 14.0.
    '\U00010D00\u200c\u088E': [(_ANY_PROFILE, 14.0)],
    '\U00010D00\u07fd\u200c\u07fd\u088E': [(_ANY_PROFILE, 14.0)],
    '\u0886\u200c\u0886': [(_ANY_PROFILE, 14.0)],
}


def _unescape(value):
    """ Unescape escapes in a python string.

    Examples:   \xFF \uFFFF \U0010FFFF \\ \n \t \r
    """
    return value.encode('ascii').decode('unicode_escape')


def main():
    """ Produce golden.json file from list of inputs.
    """

    # Read list of inputs from ASCII file, and unescape them.
    inputs = ['', '#']
    for line in sys.stdin:
        data = _unescape(line.strip())
        if not data or data[0] == '#':
            continue
        inputs.append(data)

    # For each profile, test all the inputs and produce a list of result
    # objects.
    results = []
    for profile in PROFILES:
        for data in inputs:
            unicode_version = None
            # Skip this profile for certain exceptions.
            if data in EXCEPTIONS:
                for excl in EXCEPTIONS[data]:
                    if excl[0].match(profile):
                        unicode_version = excl[1]
                        break
            try:
                output = data.encode(profile).decode('utf-8')
                reason = None
            except UnicodeEncodeError as ex:
                output = None
                reason = ex.reason
            elem = OrderedDict([('profile', profile), ('input', data),
                                ('output', output), ('error', reason)])
            if unicode_version is not None:
                elem['unicode_version'] = unicode_version
            results.append(elem)

    # Save results as an ASCII JSON file. Make sure file ends with LF.
    json.dump(results, sys.stdout, indent=2, ensure_ascii=True)
    sys.stdout.write('\n')


if __name__ == '__main__':
    main()



"""Implements the `get_profile` factory function used to create profiles."""

import precis_i18n.baseclass as _base
import precis_i18n.profile as _profile
import precis_i18n.unicode as _unicode


def _factory(profile, **kwds):
    def _construct(ucd):
        return profile(ucd, **kwds)

    return _construct


_PROFILES = {
    'identifierclass':
    _factory(_base.IdentifierClass, name='IdentifierClass'),
    'freeformclass':
    _factory(_base.FreeFormClass, name='FreeFormClass'),
    'usernamecasepreserved':
    _factory(_profile.Username, name='UsernameCasePreserved'),
    'usernamecasemapped':
    _factory(_profile.Username, name='UsernameCaseMapped', casemap='lower'),
    'usernamecasemapped_casefold':
    _factory(_profile.Username,
             name='UsernameCaseMapped:CaseFold',
             casemap='fold'),
    'usernamecasemapped_tolower':
    _factory(_profile.Username,
             name='UsernameCaseMapped:ToLower',
             casemap='lower'),
    'opaquestring':
    _factory(_profile.OpaqueString, name='OpaqueString'),
    'nicknamecasepreserved':
    _factory(_profile.Nickname, name='NicknameCasePreserved'),
    'nicknamecasemapped':
    _factory(_profile.Nickname, name='NicknameCaseMapped', casemap='lower'),
    'nicknamecasemapped_casefold':
    _factory(_profile.Nickname,
             name='NicknameCaseMapped:CaseFold',
             casemap='fold'),
    'nicknamecasemapped_tolower':
    _factory(_profile.Nickname,
             name='NicknameCaseMapped:ToLower',
             casemap='lower'),

    # Alias for backward-compatibility with previous version of codec.
    'nickname':
    _factory(_profile.Nickname, name='Nickname', casemap='lower')
}


def get_profile(name, *, unicodedata=None):
    """Return the desired PRECIS profile object.

    Choose name from:
        "IdentifierClass"
        "FreeFormClass"
        "UsernameCasePreserved"
        "UsernameCaseMapped"
        "UsernameCaseMapped:CaseFold"
        "UsernameCaseMapped:ToLower"
        "OpaqueString"
        "NicknameCasePreserved"
        "NicknameCaseMapped"
        "NicknameCaseMapped:CaseFold"
        "NicknameCaseMapped:ToLower"
        "Nickname" (alias for "NicknameCaseMapped")

    This function constructs a new profile each time; there is no cache.

    To use an alternative Unicode implementation, pass a module or object that
    implements the unicodedata interface via the unicodedata keyword argument.
    The default is to use the unicodedata module built into the Python runtime.

    Args:
        name (str): name of a PRECIS profile
        unicodedata (module|object): Alternative unicodedata interface

    Returns:
        AbstractProfile: PRECIS profile object.

    Raises:
        KeyError: Profile not found.
    """
    profile = name.lower().replace(':', '_')
    return _PROFILES[profile](_unicode.UnicodeData(unicodedata))
